/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* autoreply.c
 *
 * Copyright (C) 2019  Giam Teck Choon <choon@choon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <sys/wait.h>
#include <signal.h>
#include <bglibs/iobuf.h>
#include <bglibs/str.h>
#include <bglibs/striter.h>
#include "common.h"
#include "is_file.h"
#include "is_executable.h"
#include "str_getpath.h"

extern char *ext;
extern char *host;
extern char *sender;
extern char *recipient;
extern char *maildir;
extern char *email;

extern str tmpstr;
extern str tmp2str;
extern str tmp3str;

extern void freemem(void);

static str mstr = {0,0,0};
static str prog = {0,0,0};
static str headers = {0,0,0};

static void localfree(void)
{
	if (mstr.s) str_free(&mstr);
	if (prog.s) str_free(&prog);
	if (headers.s) str_free(&headers);
}

static void failfree(void)
{
	localfree();
	freemem();
}

/*
 * Return 1 if match otherwise return 0.
 */
static int str_filematchline(const char *filename, str *s)
{
	if (is_file(filename) && s->s) {
		str_lower(s);
		str_strip(s);
		ibuf cfn;
		if (!ibuf_open(&cfn, filename, 0)) {
			ibuf_close(&cfn);
			return 0;
		}
		static str msgstr = {0,0,0};
		while(ibuf_getstr(&cfn, &msgstr, LF)) {
			str_lower(&msgstr);
			str_strip(&msgstr);
			if (s->len >= msgstr.len && msgstr.len > 0) {
				if (strncmp(s->s, msgstr.s, msgstr.len) == 0) {
					errmsgf("{Auto-Responder skipped. "
						"Sender '}s{' listed in "
						"skip sender.\n}",
						msgstr.s);
					str_free(&msgstr);
					return 1;
				}
			}
		}
		str_free(&msgstr);
	}
	return 0;
}

static const char *skip_space(const char *s)
{
	while (*s && isspace(*s)) {
		++s;
	}
	return s;
}

/*
 * Return 1 if found header to be ignore otherwise return 0.
 */
static int check_header(const str *s, const char *h)
{
	size_t hlen = strlen(h);
	if (!strncasecmp(s->s, h, hlen)) {
		errmsgf("{Auto-Responder skipped. Message has '}s{'"
			" header.\n}", h);
		return 1;
	}
	return 0;

}

static void read_headers(void)
{
	str_truncate(&headers, 0);
	while (ibuf_getstr(&inbuf, &tmp3str, LF)) {
		if (tmp3str.s[0] == LF) {
			break;
		}
		str_cat(&headers, &tmp3str);
	}
	CATCH_NE_ZERO(lseek(STDIN_FILENO, 0L, SEEK_SET),
		      exit_cb(FAIL_SOFT, failfree));
}

/*
 * Return 1 if found in check_header to be ignore otherwise return 0.
 * References:
 * https://tools.ietf.org/html/rfc3834
 * https://github.com/jpmckinney/multi_mail/wiki/Detecting-autoresponders
 */
static int ignore_headers(void)
{
	striter i;
	striter_loop(&i, &headers, LF) {
		unsigned next = i.start + i.len + 1;
		str_truncate(&tmp3str, 0);
		str_catb(&tmp3str, headers.s + i.start, i.len + 1);
		if (next >= headers.len || !isspace(headers.s[next])) {
			if (check_header(&tmp3str, "Preference: list")) {
				return 1;
			}
			if (check_header(&tmp3str, "Preference: junk")) {
				return 1;
			}
			if (check_header(&tmp3str, "Preference: bulk")) {
				return 1;
			}
			if (check_header(&tmp3str,
					 "Auto-Submitted: auto-generated")) {
				return 1;
			}
			if (check_header(&tmp3str,
					 "Auto-Submitted: auto-replied")) {
				return 1;
			}
			if (check_header(&tmp3str, "List-ID:")) {
				return 1;
			}
			if (check_header(&tmp3str, "Mailing-List:")) {
				return 1;
			}
			if (check_header(&tmp3str, "X-Mailing-List:")) {
				return 1;
			}
			if (check_header(&tmp3str, "X-ML-Name:")) {
				return 1;
			}
			if (check_header(&tmp3str, "List-Help:")) {
				return 1;
			}
			if (check_header(&tmp3str, "List-Unsubscribe:")) {
				return 1;
			}
			if (check_header(&tmp3str, "List-Subscribe:")) {
				return 1;
			}
			if (check_header(&tmp3str, "List-Post:")) {
				return 1;
			}
			if (check_header(&tmp3str, "List-Owner:")) {
				return 1;
			}
			if (check_header(&tmp3str, "List-Archive:")) {
				return 1;
			}
			/* cPanel and BoxTrapper */
			if (check_header(&tmp3str, "Preference: auto_reply")) {
				return 1;
			}
			/* Microsoft Exchange */
			if (check_header(&tmp3str,
					 "X-Auto-Response-Suppress: All")) {
				return 1;
			}
			/* Communicate Pro */
			if (check_header(&tmp3str, "X-Autogenerated:")) {
				return 1;
			}
			/* Domain Technologie Control */
			if (check_header(&tmp3str, "X-AutoReply-From:")) {
				return 1;
			}
			if (check_header(&tmp3str, "X-Mail-Autoreply:")) {
				return 1;
			}
			/* GMail */
			if (check_header(&tmp3str, "X-Autoreply: yes")) {
				return 1;
			}
			/* Mailtraq */
			if (check_header(&tmp3str,
					 "X-POST-MessageClass: 9; "
					 "Autoresponder")) {
				return 1;
			}
		}
	}
	return 0;
}

void autoreply(void)
{
	CATCH_NE_ONE(str_copys(&mstr, sender),
		     exit_cb(FAIL_SOFT, failfree));
	CATCH_NE_ONE(str_copy2s(&tmpstr, maildir, "/autoresponse/message.txt"),
		     exit_cb(FAIL_SOFT, failfree));
	CATCH_NE_ONE(str_copy2s(&tmp2str, maildir, "/autoresponse"),
		     exit_cb(FAIL_SOFT, failfree));
	CATCH_NE_ONE(str_copy2s(&tmp3str, maildir,
				"/autoresponse/skipsender.txt"),
		     exit_cb(FAIL_SOFT, failfree));
	if (!is_file(tmpstr.s)) {
		localfree();
		return;
	}
	if (str_filematchline(tmp3str.s, &mstr)) {
		localfree();
		return;
	}
	CATCH_NE_ZERO(lseek(STDIN_FILENO, 0L, SEEK_SET),
		      exit_cb(FAIL_SOFT, failfree));
	if (!sender || !sender[0]) {
		errmsg("SENDER is empty, mail came from system account\n");
		localfree();
		return;
	}
	if (!strcmp(sender, "#@[]")) {
		errmsg("SENDER is <#@[]> (double bounce message)\n");
		localfree();
		return;
	}
	if (!strchr(sender, '@')) {
		errmsg("SENDER did not contain a hostname\n");
		localfree();
		return;
	}
	if (!strncasecmp(sender, "mailer-daemon", 13)) {
		errmsg("SENDER was mailer-daemon\n");
		localfree();
		return;
	}
	if (str_getpath("qmail-autoresponder", &prog) != 1) {
		errmsg("can't locate qmail-autoresponder\n");
		localfree();
		return;
	}
	read_headers();
	if (ignore_headers()) {
		localfree();
		return;
	}

	int fd1[2];
	CATCH_NE_ZERO(pipe(fd1), exit_cb(FAIL_SOFT, failfree));
	pid_t qapid; /* qmail-autoresponder pid */
	qapid = fork();
	switch (qapid) {
	case -1:
		errmsg("qmail-autoresponder fork() failed\n");
		failfree();
		exit(FAIL_SOFT);
	case 0: /* child process */
		setenv("ARFROM", email, 1);
		close(fd1[0]);
		CATCH_NE(dup2(fd1[1], STDOUT_FILENO), STDOUT_FILENO,
			      exit_cb(FAIL_SOFT, failfree));
		close(fd1[1]);
		const char *cmd[] = {prog.s, "-O", "no_inreplyto=1",
				     tmp2str.s, 0};
		execvp(cmd[0], (char **)cmd);
		errmsgf("{execute }s{ failed\n}", prog.s);
		failfree();
		exit(FAIL_SOFT);
	}
	/* parent process */
	close(fd1[1]);
	signal(SIGPIPE, SIG_IGN);
	int status;
	if (waitpid(qapid, &status, WUNTRACED) != qapid) {
		errmsg("System call to 'waitpid' failed\n");
		close(fd1[0]);
		failfree();
		exit(FAIL_SOFT);
	}
	if (!WIFEXITED(status)) {
		errmsgf("s{ crashed\n}", prog.s);
		close(fd1[0]);
		failfree();
		exit(FAIL_SOFT);
	}
	if (WEXITSTATUS(status)) {
		errmsgf("s{ exit status '}d{'\n}", prog.s,
			WEXITSTATUS(status));
		static str msgstr = {0,0,0};
		ibuf r;
		ibuf_init(&r, fd1[0], 0, 0, 0);
		ibuf_seek(&r, 0);
		while(ibuf_getstr(&r, &msgstr, LF)) {
			errmsg(msgstr.s);
		}
		ibuf_close(&r);
		close(fd1[0]);
		str_free(&msgstr);
		exit(FAIL_SOFT);
	}
	close(fd1[0]);
}
